@@ -12,6 +12,8 @@ from ..Exceptions import ChangesNotCommitedError
 from ..Exceptions import MergeConflict
 from ..Exceptions import NoChangesToCommitError
 from ..Exceptions import NoRemoteDefined
+from ..Exceptions import ProjectValidationError
+from ..migration import getLatestRevision
 from ..tuples.git_commit_tuple import GitCommitTuple
 
 
@@ -75,20 +77,24 @@ class GitLibMixin:
         self.__createInitialCommit()
 
     def __createInitialCommit(self):
+        # We cannot use the public `writeMetadata` method here because the
+        # branches and references to have not been set up yet
         self.__writeProjectMetadata()
         self.__writeGitIgnore()
-        self.__createCommit("HEAD", "Create default metadata.json", [])
+        self.__createCommit(
+            "HEAD", "Create default metadata.json", [], self.__index
+        )
 
     def __writeProjectMetadata(self):
         # Create default metadata.json
         data = json.dumps(
-            {},
+            {"revision": getLatestRevision()},
             indent=4,
             sort_keys=True,
             separators=(", ", ": "),
         ).encode()
         path = "metadata.json"
-        # Write the file, with out preparing the working branch
+        # Write the file, *without* preparing the working branch
         self.__index.add(
             pygit2.IndexEntry(
                 path, self.__repo.create_blob(data), pygit2.GIT_FILEMODE_BLOB
@@ -107,6 +113,29 @@ class GitLibMixin:
         )
         self.__isIndexDirty = True
 
+    @property
+    def metadata(self):
+        tree = self.__workingTree
+        data = (tree / "metadata.json").data
+        metadataStr = data.decode()
+        metadata = json.loads(metadataStr)
+        return metadata
+
+    def writeMetadata(self, metadata):
+        oid = self.__repo.create_blob(
+            json.dumps(
+                metadata,
+                indent=4,
+                sort_keys=True,
+                separators=(", ", ": "),
+            ).encode()
+        )
+        self.__prepareWorkingBranch()
+        self.__index.add(
+            pygit2.IndexEntry("metadata.json", oid, pygit2.GIT_FILEMODE_BLOB)
+        )
+        self.__isIndexDirty = True
+
     @classmethod
     def cloneGitRepo(
         cls,
@@ -245,19 +274,21 @@ class GitLibMixin:
             descendents = [descendentCommit.id]
         else:
             descendents = []
-        self.__createCommit(self.__workingBranch.name, msg, descendents)
+        self.__createCommit(
+            self.__workingBranch.name, msg, descendents, self.__index
+        )
 
     def __createCommit(
-        self, branchName: str, msg: str, descendents: list[str]
+        self, targetName: str, msg: str, descendents: list[str], index
     ) -> str:
 
-        tree = self.__index.write_tree()
+        tree = index.write_tree()
         author = self.__committer if self.__committer else self.defaultCommitter
 
         self.__isIndexDirty = False
 
         return self.__repo.create_commit(
-            branchName,
+            targetName,
             author,
             author,
             msg,
@@ -411,6 +442,7 @@ class GitLibMixin:
             makeRef(ATTUNE_WORKING_BRANCH),
             mergeCommitMessage,
             [checkedLastCommit.id],
+            self.__index,
         )
 
         """
@@ -506,6 +538,8 @@ class GitLibMixin:
         working = self.__workingBranch
         checked = self.__checkedOutBranch
 
+        logger.debug(f"Merging branches {sourceBranch} to {checked.name}")
+
         if working.target != checked.target:
             raise RuntimeError(
                 f"Squash the commits for {self.__checkedOutBranch} first"
@@ -516,26 +550,58 @@ class GitLibMixin:
 
         mergeAnalysis, _ = self.__repo.merge_analysis(sourceBranchId)
         if mergeAnalysis & pygit2.GIT_MERGE_ANALYSIS_UP_TO_DATE:
+            logger.debug("Both branches are up-to-date. Nothing to merge")
             return
 
         if mergeAnalysis & pygit2.GIT_MERGE_ANALYSIS_FASTFORWARD:
-            # Fast forward checked branch to source branch
+            logger.debug("Branch can be fast-forwarded")
+            # Fast-forward checked branch to source branch
             checked.set_target(sourceBranchId)
 
         elif mergeAnalysis & pygit2.GIT_MERGE_ANALYSIS_NORMAL:
+            """
+            IMPORTANT!!
+            The `merge` method on the `Repository` object and the merging
+            procedure below is dependent on the state of the index and the
+            working tree. We cannot use self.__index or any other methods
+            that mutate the index or working tree here. self.__index
+            reloads the index if there is no working branch which causes
+            the merged index to be lost and the merged changes won't be
+            committed.
+            """
+
+            logger.debug("Merging branches")
             self.__repo.merge(sourceBranchId)
             # Check for merge conflicts and reverse merge if there are any
-            if self.__index.conflicts:
+            if self.__repo.index.conflicts:
                 self.__repo.reset(checkedBranchId, pygit2.GIT_RESET_HARD)
                 raise MergeConflict("Cannot complete merge due to conflicts")
 
+            logger.debug("Branches merged. Creating merge commit")
             self.__createCommit(
-                checked.name,
-                "Merge",
-                [checkedBranchId, sourceBranchId],
+                "HEAD",
+                "Merged pulled changes",
+                [self.__repo.head.target, sourceBranchId],
+                self.__repo.index,
             )
+
             # This is so Git CLI does not show status as MERGING
             self.__repo.state_cleanup()
 
         else:
             raise RuntimeError("Unknown Git merge analysis result")
+
+        # Check for project validation errors after merge
+        try:
+            # This is the load method from GitObjectStorageContext
+            self.load()
+        except ProjectValidationError:
+            logger.debug(
+                "Merged project failed validation. Resetting to "
+                "previous state"
+            )
+            self.__repo.reset(checkedBranchId, pygit2.GIT_RESET_HARD)
+            logger.debug("Project reset")
+            raise Exception("Merging failed due to project validation errors")
+
+        logger.debug("No validation errors after merge")
