@@ -16,9 +16,8 @@ logger = logging.getLogger(__name__)
 def runMigrationsForStorageContext(storageContext: ObjectStorageContext):
     currentRev = storageContext.getRevision()
 
-    # Alembic requires a DB to read the version from. We mock an in-memory DB
-    # that has a table `alembic_version` with a single value as the current rev
-    db = sqlite3.connect("file:")
+    # Create a temporary database file
+    db = sqlite3.connect("")
     cur = db.cursor()
     cur.execute("CREATE TABLE alembic_version(version_num varchar)")
     # What if someone sets currentRev to "Robert'); DROP TABLE users;--" ?
